#!/bin/perl
# Unix Field Pack for EPAgent
# Copyright 2004-2006 Wily Technology, Inc. All rights reserved.
# Version 0.3 - 29 December 2006
# Sylvain Gibassier

# Note: This script creates a metric called "vmstat:Success"
# The value for this metric is 2 is the script is functioning normally.
# If there is a problem, the value of this metric should be 1 and there
# should be a vmstat_errors.log file in the logs directory. This file contains
# some diagnostic information for the problem that this script is experiencing.

use Sys::Hostname;
# warnings and strict commented out for the default version of Perl that comes with Solaris 8
# use warnings;
# use strict;

my $WILYHOME="/apps/websphere6/epagent";
my $error_log = "$WILYHOME/logs/vmstat_errors.log";

sub identifyColumns {
	my %headerMappings = (
		# Kernel Threads
		'r' => {'metricName'=>'Kernel Threads:Runnable','metricType'=>'IntCounter','ratio'=>1},
		'b' => {'metricName'=>'Kernel Threads:In Wait Queue','metricType'=>'IntCounter','ratio'=>1},
		'w' => {'metricName'=>'Kernel Threads:Swapped','metricType'=>'IntCounter','ratio'=>1},

		# Memory
		'avm' => {'metricName'=>'Memory:Active Virtual Pages (Bytes)','metricType'=>'LongCounter','ratio'=>4096},
		'fre' => {'metricName'=>'Memory:Free List (Bytes)','metricType'=>'LongCounter','ratio'=>4096},
		'free' => {'metricName'=>'Memory:Free List (Bytes)','metricType'=>'LongCounter','ratio'=>1024},
		'cache' => {'metricName'=>'Memory:Cache (Bytes)','metricType'=>'LongCounter','ratio'=>1024},
		'buff' => {'metricName'=>'Memory:Buffers (Bytes)','metricType'=>'LongCounter','ratio'=>1024},
		'sr' => {'metricName'=>'Memory:Scan Rate','metricType'=>'IntCounter','ratio'=>1},
		
		# swap
		'swap' => {'metricName'=>'Swap:Free (Bytes)','metricType'=>'LongCounter','ratio'=>1024},
		'swpd' => {'metricName'=>'Swap:Active (Bytes)','metricType'=>'LongCounter','ratio'=>1024},
		'pi' => {'metricName'=>'Swap:Pages paged in from paging space (Bytes/s)','metricType'=>'IntCounter','ratio'=>1024},
		'po' => {'metricName'=>'Swap:Pages paged out to paging space (Bytes/s)','metricType'=>'IntCounter','ratio'=>1024},
		'si' => {'metricName'=>'Swap:Pages paged in from paging space (Bytes/s)','metricType'=>'IntCounter','ratio'=>1024},
		'so' => {'metricName'=>'Swap:Pages paged out to paging space (Bytes/s)','metricType'=>'IntCounter','ratio'=>1024},
		
		# system
		'in' => {'metricName'=>'System:Device Interrupts','metricType'=>'LongCounter','ratio'=>1},
		# there is another sy
		'systemCalls' => {'metricName'=>'System:System Calls','metricType'=>'LongCounter','ratio'=>1},
		'cs' => {'metricName'=>'System:Context Switches','metricType'=>'LongCounter','ratio'=>1},

		# cpu utilization
		'us' => {'metricName'=>'Cpu:User %','metricType'=>'IntCounter','ratio'=>1},
		# there is another sy
		'systemCpu' => {'metricName'=>'Cpu:System %','metricType'=>'IntCounter','ratio'=>1},
		'wa' => {'metricName'=>'Cpu:IOWait %','metricType'=>'IntCounter','ratio'=>1},
		'id' => {'metricName'=>'Cpu:Idle %','metricType'=>'IntCounter','ratio'=>1},
	);
	
	if ($^O eq 'aix') {
		$headerMappings{'pi'}->{'ratio'}=4096;
		$headerMappings{'po'}->{'ratio'}=4096;
	}
	
	if ($^O eq 'hpux') {
		$headerMappings{'pi'}->{'ratio'}=4096;
		$headerMappings{'po'}->{'ratio'}=4096;
		$headerMappings{'free'}->{'ratio'}=4096;
	}

	if ($^O eq 'linux') {
		$headerMappings{'free'}->{'metricName'}='Memory:Free Physical Memory (Bytes)';
	}
	
	my $precedent='';
	my @vmstatConfiguration;
	foreach (@_) {
		if ($_ ne 'sy') {
			push @vmstatConfiguration,$headerMappings{$_};
		} else {
			if ($precedent eq 'us') {
				push @vmstatConfiguration,$headerMappings{'systemCpu'};
			} else {
				push @vmstatConfiguration,$headerMappings{'systemCalls'};
			}		
		}
		$precedent = $_;
	}
	return @vmstatConfiguration;
}

sub printColumnIdentifiers {
	my $count = 0;
	foreach (@_) {
		if ($_) {
			print "$count: $_->{'metricName'}\n";
		} else {
			print "$count: null\n";
		}
		$count++;
	}
}

sub getPhysicalMemorySize {
	my $physicalMemorySize = -1;

	if ($^O eq 'aix' || $^O eq 'solaris') {
		$physicalMemorySize = getPhysicalMemorySize4SolarisOrAIX();
	} elsif ($^O eq 'linux') {
		$physicalMemorySize = getPhysicalMemorySize4Linux();
	} elsif ($^O eq 'MSWin32') {
		$physicalMemorySize = 1073741824;
	}

	return $physicalMemorySize;
}

sub getPhysicalMemorySize4SolarisOrAIX {
	my $physicalMemorySize = -1;
	eval {
		my $prtconf_command = 'prtconf 2>/dev/null |';
		
		if (scalar(@_) > 0) {
			$prtconf_command = $_[0];
		}

		my $prtconf_process_id = open(PRTCONF,$prtconf_command) || die "Error when executing \"$prtconf_command\": $!";
		
		while(<PRTCONF>) {
			chomp;
			if (/^Memory Size:\s*(\d+)\s*M(B|egabytes)$/i) {
				$physicalMemorySize = $1 * 1048576;
				last;
			}
		}

		close(PRTCONF);
		kill 15,$prtconf_process_id;
	};

	return $physicalMemorySize;
}

sub getPhysicalMemorySize4Linux {
	my $physicalMemorySize = -1;
	eval {
		my $free_command = 'free|';
		
		if (scalar(@_) > 0) {
			$free_command = $_[0];
		}

		open(FREE,$free_command) || die "Error when executing \"$free_command\": $!";
		
		while(<FREE>) {
			chomp;
			if (/^Mem\w*:\s*(\d*)/i) {
				$physicalMemorySize = $1 * 1024;
			}
		}

		close(FREE);
	};

	return $physicalMemorySize;
}

eval {
	my $physicalMemorySize = getPhysicalMemorySize(); 

	my $firstTime = 1;

	my $child_process_id = 0;
	my $vmstat_command = 'vmstat 7|';
	my $host = hostname;
	my $vmstat_pid_file_name="$WILYHOME/pids/VMSTAT.${host}.pid";
	$error_log = "$WILYHOME/logs/vmstat_errors.${host}.log";

	if ($^O eq 'MSWin32') {
		$vmstat_command = '../vmstat_test.txt';
	}

	$child_process_id = open(VMSTAT, $vmstat_command) || die "Error when executing \"$vmstat_command\": $!";

	# write the child process id to file
	open(PROCESS_ID,">>$vmstat_pid_file_name") || die "Error when opening file \"$vmstat_pid_file_name\": $!";
	print PROCESS_ID "$child_process_id ";
	close(PROCESS_ID);

	my @columnIdentifiers;

	while (<VMSTAT>) {
		chomp;
		my @list = split;

		if (scalar(@list) > 6) {
			if (scalar(@list) == map(/[\d.]+/,@list)) {
				# ignore first line of vmstat
				if ($firstTime) {
					$firstTime = 0;
				} else {
					for(my $i=0;$i<scalar(@list);$i++) {
						# print only the columns we know about
						if ($columnIdentifiers[$i]) {
							print qq{<metric type="$columnIdentifiers[$i]->{'metricType'}" name="vmstat|$columnIdentifiers[$i]->{'metricName'}" value="} . $list[$i]*$columnIdentifiers[$i]->{'ratio'} . qq{"/>\n};
							
							# print cpu utilization metric
							if ($columnIdentifiers[$i]->{'metricName'} eq 'Cpu:Idle %') {
								print qq{<metric type="IntCounter" name="vmstat|Cpu:Utilization %" value="} . (100 - ($list[$i]*$columnIdentifiers[$i]->{'ratio'})) . qq{"/>\n};
							}

							# memory percentage
							if ($columnIdentifiers[$i]->{'metricName'} =~ /^Memory:.*\(Bytes\)$/ && ($physicalMemorySize > 0)) {
								my $percentage = (100 * $list[$i] * $columnIdentifiers[$i]->{'ratio'} / $physicalMemorySize);
								my $percentageMetricName = $columnIdentifiers[$i]->{'metricName'};
								$percentageMetricName =~ s/Bytes/% of physical memory/;
								printf qq{<metric type="IntCounter" name="vmstat|%s" value="%d"/>\n},$percentageMetricName,$percentage;
							}
						}
					}
					
					if ($physicalMemorySize > 0) {
						print qq{<metric type="LongCounter" name="vmstat|Memory:Physical (Bytes)" value="$physicalMemorySize"/>\n};
					}

					# gives the end user the opportunity to monitor the vmstat plugin
					print qq{<metric type="IntCounter" name="vmstat:Success" value="2"/>\n};
				}

			} elsif ($list[0] eq 'r' && $list[1] eq 'b') {
				# this is the headers, parse them to identify the columns
				@columnIdentifiers = identifyColumns(@list);	
				#printColumnIdentifiers @columnIdentifiers;
			}
		}
	}
	close(VMSTAT);
};

if ( $@ ) {
	# write the error message to an error log for troubleshooting	
	open(ERROR_LOG,">>$error_log") || die "Error when opening file \"$error_log\": $!";
	print ERROR_LOG localtime() . " - " . $@;
	close(ERROR_LOG);

	while (1) {
		# gives the end user the opportunity to monitor the vmstat plugin
		print qq{<metric type="IntCounter" name="vmstat:Success" value="1"/>\n};
		# we don't want to return here because EPAgent is going to try to restart the plugin
		# and it is very unlikely that this will fix the problem we are experiencing
		sleep 15;
	}
}
